@@ -1,5 +1,5 @@
 // L2TPNS Global Stuff
-// $Id: l2tpns.h,v 1.113.2.3 2006/12/02 14:09:14 bodea Exp $
+// $Id: l2tpns.h,v 1.121 2009-12-08 14:49:28 bodea Exp $
 
 #ifndef __L2TPNS_H__
 #define __L2TPNS_H__
@@ -9,18 +9,25 @@
 #include <signal.h>
 #include <stdlib.h>
 #include <netinet/in.h>
+#include <net/ethernet.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
 #include <sys/types.h>
 #include <libcli.h>
 
-#define VERSION	"2.1.21"
+#define VERSION        "2.3.1"
 
 // Limits
 #define MAXTUNNEL	500		// could be up to 65535
+#define MAXBUNDLE	300		// could be up to 65535
+#define MAXBUNDLESES	12		// Maximum number of member links in bundle
+#define MAXADDRESS	20		// Maximum length for the Endpoint Discrminiator address
 #define MAXSESSION	60000		// could be up to 65535
 #define MAXTBFS		6000		// Maximum token bucket filters. Might need up to 2 * session.
 
+// Tunnel Id reserved for pppoe
+#define TUNNEL_ID_PPPOE	1
+
 #define RADIUS_SHIFT	6
 #define RADIUS_FDS	(1 << RADIUS_SHIFT)
 #define RADIUS_MASK	((1 << RADIUS_SHIFT) - 1)
@@ -35,18 +42,40 @@
 #define PPPoE_MRU	1492		// maximum PPPoE MRU (rfc2516: 1500 less PPPoE header (6) and PPP protocol ID (2))
 #define MAXETHER	(MAXMTU+18)	// max packet we try sending to tun
 #define MAXTEL		96		// telephone number
+#define MAXHOSTNAME	256		// hostname
 #define MAXUSER		128		// username
 #define MAXPASS		128		// password
+#define MAXCLASS	128		// radius class attribute size
 #define MAXPLUGINS	20		// maximum number of plugins to load
 #define MAXRADSERVER	10		// max radius servers
 #define MAXROUTE	10		// max static routes per session
+#define MAXROUTE6	5		// max static Ipv6 routes per session
 #define MAXIPPOOL	131072		// max number of ip addresses in pool
 #define RINGBUFFER_SIZE	10000		// Number of ringbuffer entries to allocate
 #define MAX_LOG_LENGTH	512		// Maximum size of log message
-#define ECHO_TIMEOUT	60		// Time between last packet sent and LCP ECHO generation
-#define IDLE_TIMEOUT	240		// Time between last packet sent and LCP ECHO generation
+#define ECHO_TIMEOUT	10		// Time between last packet sent and LCP ECHO generation
+#define IDLE_ECHO_TIMEOUT	240		// Time between last packet seen and session shutdown
 #define BUSY_WAIT_TIME	3000		// 5 minutes in 1/10th seconds to wait for radius to cleanup on shutdown
 
+#define MP_BEGIN        0x80            // This value is used when (b)egin bit is set in MP header
+#define MP_END          0x40            // This value is used when (e)nd bit is set in MP header
+#define MP_BOTH_BITS    0xC0            // This value is used when both bits (begin and end) are set in MP header
+
+#define MINFRAGLEN	64		// minimum fragment length
+#define MAXFRAGLEN	1496	// Maximum length for Multilink fragment (The multilink may contain only one link)
+#define MAXFRAGNUM	512		// Maximum number of Multilink fragment in a bundle (must be in the form of 2^X)
+					// it's not expected to have a space for more than 10 unassembled packets = 10 * MAXBUNDLESES
+#define	MAXFRAGNUM_MASK	(MAXFRAGNUM - 1)		// Must be equal to MAXFRAGNUM-1
+
+// Multi bind address constants
+#define MAX_UDPFD 4
+#define MAX_BINDADDR MAX_UDPFD
+// + 1 for the LAC Hostname
+#define MAX_NBHOSTNAME (MAX_UDPFD + 1)
+// 4 + 1 for the udplac
+#define INIT_TABUDPFD {-1, -1, -1, -1, -1}
+#define INIT_TABUDPVAR {0, 0, 0, 0, 0}
+
 // Constants
 #ifndef ETCDIR
 #define ETCDIR		"/etc/l2tpns"
@@ -186,6 +215,7 @@ enum {
 
 // Types
 typedef uint16_t sessionidt;
+typedef uint16_t bundleidt;
 typedef uint16_t tunnelidt;
 typedef uint32_t clockt;
 typedef uint8_t hasht[16];
@@ -219,10 +249,18 @@ struct cli_tunnel_actions {
 typedef struct			// route
 {
 	in_addr_t ip;
-	in_addr_t mask;
+	int prefixlen;
 }
 routet;
 
+// structures
+typedef struct			// route
+{
+	struct in6_addr ipv6route;	// Static IPv6 route
+	uint8_t ipv6prefixlen;		// IPv6 route prefix length
+}
+routet6;
+
 typedef struct controls		// control message
 {
 	struct controls *next;	// next in queue
@@ -231,6 +269,22 @@ typedef struct controls		// control message
 }
 controlt;
 
+typedef struct {
+	uint8_t length;			// Endpoint Discriminator length
+	uint8_t addr_class;		// Endpoint Discriminator class
+	uint8_t address[MAXADDRESS];	// Endpoint Discriminator address
+} epdist;
+
+typedef struct {
+	sessionidt sid;			// Fragment originating session
+	tunnelidt tid;			// Fragment originating tunnel
+	uint8_t	flags;			// MP frame flags
+	uint32_t seq;			// fragment seq num
+	uint32_t jitteravg;
+        uint16_t length;                // Fragment length
+        uint8_t data[MAXFRAGLEN];       // Fragment data
+} fragmentt;
+
 typedef struct
 {
 	sessionidt next;		// next session in linked list
@@ -244,11 +298,10 @@ typedef struct
 		uint8_t ipv6cp:4;	//   IPV6CP state
 		uint8_t ccp:4;		//   CCP    state
 	} ppp;
-	char reserved_1[2];		// unused: padding
+	uint16_t mru;			// maximum receive unit
 	in_addr_t ip;			// IP of session set by RADIUS response (host byte order).
 	int ip_pool_index;		// index to IP pool
 	uint32_t unique_id;		// unique session id
-	char reserved_2[4];		// unused: was ns/nr
 	uint32_t magic;			// ppp magic number
 	uint32_t pin, pout;		// packet counts
 	uint32_t cin, cout;		// byte counts
@@ -258,30 +311,74 @@ typedef struct
 	uint16_t throttle_out;		// downstream throttle rate
 	uint8_t filter_in;		// input filter index (to ip_filters[N-1]; 0 if none)
 	uint8_t filter_out;		// output filter index
-	uint16_t mru;			// maximum receive unit
+	uint16_t snoop_port;		// Interception destination port
+	in_addr_t snoop_ip;		// Interception destination IP
 	clockt opened;			// when started
 	clockt die;			// being closed, when to finally free
+	uint32_t session_timeout;	// Maximum session time in seconds
+	uint32_t idle_timeout;		// Maximum idle time in seconds
 	time_t last_packet;		// Last packet from the user (used for idle timeouts)
+	time_t last_data;		// Last data packet to/from the user (used for idle timeouts)
 	in_addr_t dns1, dns2;		// DNS servers
 	routet route[MAXROUTE];		// static routes
 	uint16_t tbf_in;		// filter bucket for throttling in from the user.
 	uint16_t tbf_out;		// filter bucket for throttling out to the user.
 	int random_vector_length;
 	uint8_t random_vector[MAXTEL];
-	char user[MAXUSER];		// user (needed in seesion for radius stop messages)
+	char user[MAXUSER];		// user (needed in session for radius stop messages)
 	char called[MAXTEL];		// called number
 	char calling[MAXTEL];		// calling number
 	uint32_t tx_connect_speed;
 	uint32_t rx_connect_speed;
-	in_addr_t snoop_ip;		// Interception destination IP
-	uint16_t snoop_port;		// Interception destination port
+	clockt timeout;                 // Session timeout
+	uint32_t mrru;                  // Multilink Max-Receive-Reconstructed-Unit
+	epdist epdis;                   // Multilink Endpoint Discriminator
+	bundleidt bundle;               // Multilink Bundle Identifier
+	uint8_t mssf;                   // Multilink Short Sequence Number Header Format
 	uint8_t walled_garden;		// is this session gardened?
-	uint8_t ipv6prefixlen;		// IPv6 route prefix length
-	struct in6_addr ipv6route;	// Static IPv6 route
-	char reserved_3[11];		// Space to expand structure without changing HB_VERSION
+	uint8_t classlen;		// class (needed for radius accounting messages)
+	char class[MAXCLASS];
+	sessionidt forwardtosession;	// LNS id_session to forward
+	uint8_t src_hwaddr[ETH_ALEN];	// MAC addr source (for pppoe sessions 6 bytes)
+	uint32_t dhcpv6_prefix_iaid;	// prefix iaid requested by client
+	uint32_t dhcpv6_iana_iaid;		// iaid of iana requested by client
+	struct in6_addr ipv6address;	// Framed Ipv6 address
+	struct dhcp6_opt_clientid dhcpv6_client_id; // Size max (headers + DUID)
+	routet6 route6[MAXROUTE6];		// static IPv6 routes
+	char reserved[4];		// Space to expand structure without changing HB_VERSION
 }
 sessiont;
 
+typedef struct
+{
+        int state;                              // current state (bundlestate enum)
+        uint32_t seq_num_t;                     // Sequence Number (transmission)
+        uint32_t timeout;                       // Session-Timeout for bundle
+	uint32_t max_seq;			// Max value of sequence number field
+        uint8_t num_of_links;                   // Number of links joint to this bundle
+        uint32_t online_time;                   // The time this bundle is online
+        clockt last_check;                      // Last time the timeout is checked
+        uint32_t mrru;                          // Multilink Max-Receive-Reconstructed-Unit
+        uint8_t mssf;                           // Multilink Short Sequence Number Header Format
+        epdist epdis;                           // Multilink Endpoint Discriminator
+        char user[MAXUSER];                     // Needed for matching member links
+        sessionidt current_ses;                 // Current session to use for sending (used in RR load-balancing)
+        sessionidt members[MAXBUNDLESES];       // Array for member links sessions
+}
+bundlet;
+
+typedef struct
+{
+	fragmentt fragment[MAXFRAGNUM];
+	uint8_t reassembled_frame[MAXETHER];    // The reassembled frame
+	uint16_t re_frame_len;                  // The reassembled frame length
+	uint16_t re_frame_begin_index, re_frame_end_index;	// reassembled frame begin index, end index respectively
+	uint16_t start_index, end_index;	// start and end sequence numbers available on the fragments array respectively
+	uint32_t M;				// minimum frame sequence number received over all bundle members
+	uint32_t start_seq;                     // Last received frame sequence number (bearing B bit)
+}
+fragmentationt;
+
 #define AUTHPAP		1	// allow PAP
 #define AUTHCHAP	2	// allow CHAP
 
@@ -311,6 +408,15 @@ typedef struct
 	// our MRU
 	uint16_t ppp_mru;
 
+	// our MRRU
+	uint16_t mp_mrru;
+
+	// our mssf
+	uint16_t mp_mssf;
+
+	// our Endpoint Discriminator
+	in_addr_t mp_epdis;
+
 	// DoS prevention
 	clockt last_packet_out;
 	uint32_t packets_out;
@@ -324,6 +430,14 @@ typedef struct
 
 	// last LCP Echo
 	time_t last_echo;
+
+	// Last Multilink frame sequence number received
+	uint32_t last_seq;
+
+	// jitter average of the session
+	uint32_t jitteravg;
+	// time in milliseconds of the last fragment.
+	uint64_t prev_time;
 } sessionlocalt;
 
 // session flags
@@ -331,7 +445,7 @@ typedef struct
 #define SESSION_ACFC	(1 << 1)	// use Address-and-Control-Field-Compression
 #define SESSION_STARTED	(1 << 2)	// RADIUS Start record sent
 
-// 168 bytes per tunnel
+// 328 bytes per tunnel
 typedef struct
 {
 	tunnelidt far;		// far end tunnel ID
@@ -351,6 +465,9 @@ typedef struct
 	uint16_t controlc;	// outstaind messages in queue
 	controlt *controls;	// oldest message
 	controlt *controle;	// newest message
+	uint16_t isremotelns;	// != 0 if the tunnel is to remote LNS (== index on the conf remote lns)
+	uint16_t indexudp;	// Index UDP file handle (in udpfd[])
+	char reserved[12];		// Space to expand structure without changing HB_VERSION
 }
 tunnelt;
 
@@ -407,6 +524,23 @@ enum
 	TUNNELUNDEF,		// Undefined
 };
 
+enum
+{
+	BUNDLEFREE,		// Not in use
+	BUNDLEOPEN,		// Active bundle
+	BUNDLEUNDEF,		// Undefined
+};
+
+enum
+{
+	NULLCLASS = 0,		//End Point Discriminator classes
+	LOCALADDR,
+	IPADDR,
+	IEEEMACADDR,
+	PPPMAGIC,
+	PSNDN,
+};
+
 enum
 {
 	RADIUSNULL,             // Not in use
@@ -416,6 +550,7 @@ enum
 	RADIUSSTOP,             // sending stop accounting to RADIUS server
 	RADIUSINTERIM,		// sending interim accounting to RADIUS server
 	RADIUSWAIT,		// waiting timeout before available, in case delayed replies
+	RADIUSJUSTAUTH,         // sending auth to RADIUS server, just authentication, no ip assigning
 };
 
 struct Tstats
@@ -471,6 +606,7 @@ struct Tstats
     uint32_t	call_processudp;
     uint32_t	call_sessionbyip;
     uint32_t	call_sessionbyipv6;
+    uint32_t	call_sessionbyipv6new;
     uint32_t	call_sessionbyuser;
     uint32_t	call_sendarp;
     uint32_t	call_sendipcp;
@@ -497,6 +633,7 @@ struct Tstats
     uint32_t	call_radiussend;
     uint32_t	call_radiusretry;
     uint32_t    call_random_data;
+    uint32_t    call_dhcpv6_process;
 #endif
 };
 
@@ -520,6 +657,10 @@ struct Tstats
 #define SET_STAT(x, y)
 #endif
 
+#ifndef IFNAMSIZ
+# define IFNAMSIZ 16
+#endif
+
 typedef struct
 {
 	int		debug;				// debugging level
@@ -535,7 +676,7 @@ typedef struct
 	int		reload_config;			// flag to re-read config (set by cli)
 	int		multi_read_count;		// amount of packets to read per fd in processing loop
 
-	char		tundevice[10];			// tun device name
+	char		tundevicename[IFNAMSIZ];		// tun device name
 	char		log_filename[128];
 
 	char		l2tp_secret[64];		// L2TP shared secret
@@ -546,6 +687,7 @@ typedef struct
 	int		ppp_restart_time;		// timeout for PPP restart
 	int		ppp_max_configure;		// max lcp configure requests to send
 	int		ppp_max_failure;		// max lcp configure naks to send
+	int		ppp_keepalive;			// send echoes regardless
 
 	char		radiussecret[64];
 	int		radius_accounting;
@@ -554,13 +696,16 @@ typedef struct
 	uint16_t	radiusport[MAXRADSERVER];	// radius base ports
 	uint8_t		numradiusservers;		// radius server count
 
-	uint16_t	radius_dae_port;		// local port for radius dae
+	uint16_t	radius_dae_port;		// port for radius DAE
+	uint16_t	radius_bind_min;		// port range for udp sockets used to send/recv radius packets
+	uint16_t	radius_bind_max;
 
 	char		radius_authtypes_s[32];		// list of valid authentication types (chap, pap) in order of preference
 	int		radius_authtypes;
 	int		radius_authprefer;
 
 	int		allow_duplicate_users;		// allow multiple logins with the same username
+	int		kill_timedout_sessions;		// kill authenticated sessions with "session_timeout == 0"
 
 	in_addr_t	default_dns1, default_dns2;
 
@@ -568,11 +713,11 @@ typedef struct
 	int		num_tbfs;			// number of throttle buckets
 
 	char		accounting_dir[128];
+	int		account_all_origin; // Accouting all origin (LAC data + Remote LNS Data + PPPOE data)
 	in_addr_t	bind_address;
 	in_addr_t	peer_address;
 	int		send_garp;			// Set to true to garp for vip address on startup
 
-	int		target_uid;
 	int		dump_speed;
 	char		plugins[64][MAXPLUGINS];
 	char		old_plugins[64][MAXPLUGINS];
@@ -583,8 +728,10 @@ typedef struct
 	int		lock_pages;			// Lock pages into memory.
 	int		icmp_rate;			// Max number of ICMP unreachable per second to send
 	int		max_packets;			// DoS prevention: per session limit of packets/0.1s
+	char		epdis_addr[20];			// MP Endpoint Discriminator address
 
 	in_addr_t	cluster_address;		// Multicast address of cluster.
+	int			cluster_port;			// UDP port of cluster.
 							// Send to this address to have everyone hear.
 	char		cluster_interface[64];		// Which interface to listen for multicast on.
 	int		cluster_iam_master;		// Are we the cluster master???
@@ -594,8 +741,10 @@ typedef struct
 	int		cluster_seq_number;		// Sequence number of the next heartbeat we'll send out
 							// (or the seq number we're next expecting if we're a slave).
 	int		cluster_undefined_sessions;	// How many sessions we're yet to receive from the master.
+	int		cluster_undefined_bundles;	// How many bundles we're yet to receive from the master.
 	int		cluster_undefined_tunnels;	// How many tunnels we're yet to receive from the master.
 	int		cluster_highest_sessionid;
+	int		cluster_highest_bundleid;
 	int		cluster_highest_tunnelid;
 	clockt		cluster_last_hb;		// Last time we saw a heartbeat from the master.
 	int		cluster_last_hb_ver;		// Heartbeat version last seen from master
@@ -611,6 +760,10 @@ typedef struct
 
 	int		cluster_master_min_adv;		// Master advertises routes while the number of up to date
 							// slaves is less than this value.
+	in_addr_t	cli_bind_address;		// bind address for CLI
+	char		hostname[MAXHOSTNAME];	// hostname (overridden by -h on command line)
+	// Guest change
+	char            guest_user[MAXUSER];            // Guest account username
 
 #ifdef BGP
 #define BGP_NUM_PEERS	2
@@ -620,8 +773,45 @@ typedef struct
 	    	uint16_t as;
 		int keepalive;
 		int hold;
+		struct in_addr update_source;
 	} neighbour[BGP_NUM_PEERS];
+	in_addr_t	nexthop_address;
+	struct in6_addr	nexthop6_address;
 #endif
+
+	int echo_timeout; // Time between last packet sent and LCP ECHO generation
+	int idle_echo_timeout; // Time between last packet seen and
+						   // Drop sessions who have not responded within IDLE_ECHO_TIMEOUT seconds
+	in_addr_t iftun_address;
+	int disable_lac_func;
+	int auth_tunnel_change_addr_src;
+	int highest_rlnsid;
+	uint16_t bind_portremotelns;
+	in_addr_t bind_address_remotelns;
+	char pppoe_if_to_bind[IFNAMSIZ];	// Name pppoe interface to bind
+	char pppoe_service_name[64];	// pppoe service name
+	char pppoe_ac_name[64];
+	uint8_t pppoe_hwaddr[ETH_ALEN];	// MAC addr of interface pppoe to bind
+	int pppoe_only_equal_svc_name; // Accept only PADI with service-name equal to server
+	int disable_sending_hello; // Disable l2tp sending HELLO message for Apple compatibility.
+	int disable_no_spoof; // Disable no spoof (permit load balancing client --> internet)
+	int nbudpfd; // number UDP file handle
+	int nbmultiaddress; // number multi address to bind
+	int indexlacudpfd;	// Index UDP LAC file handle (in udpfd[])
+	int nbmultihostname;	// number hostname, normally the same number as the nbudpfd
+	int no_throttle_local_IP;	// no throttle traffic from session to session
+	in_addr_t bind_n_address[MAX_BINDADDR];
+	in_addr_t iftun_n_address[MAX_BINDADDR];
+	char bind_multi_address[256];
+	char multi_hostname[512];
+	char multi_n_hostname[MAX_NBHOSTNAME][MAXHOSTNAME];	// list hostname
+	struct in6_addr default_ipv6_dns1;
+	struct in6_addr default_ipv6_dns2;
+	uint32_t dhcp6_preferred_lifetime;		// preferred lifetime (see rfc3315)
+	uint32_t dhcp6_valid_lifetime;		// valid lifetime (see rfc3315)
+	uint32_t dhcp6_server_duid;		// DUID of dhcpv6 server (see rfc3315)
+	uint32_t dns6_lifetime;		// RDNSS lifetime default 1200 (see rfc6106, rfc4861) (MaxRtrAdvInterval <= Lifetime <= 2*MaxRtrAdvInterval)
+	char default_ipv6_domain_list[255];
 } configt;
 
 enum config_typet { INT, STRING, UNSIGNED_LONG, SHORT, BOOL, IPv4, IPv6 };
@@ -715,6 +905,13 @@ typedef struct
 #define TERM_PORT_REINIT		21
 #define TERM_PORT_DISABLED		22
 
+// on slaves, alow BGP to withdraw cleanly before exiting
+#define QUIT_DELAY	5
+
+// quit actions (master)
+#define QUIT_FAILOVER	1 // SIGTERM: exit when all control messages have been acked (for cluster failover)
+#define QUIT_SHUTDOWN	2 // SIGQUIT: shutdown sessions/tunnels, reject new connections
+
 // arp.c
 void sendarp(int ifr_idx, const unsigned char* mac, in_addr_t ip);
 
@@ -723,18 +920,23 @@ void sendarp(int ifr_idx, const unsigned char* mac, in_addr_t ip);
 void processpap(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l);
 void processchap(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l);
 void lcp_open(sessionidt s, tunnelidt t);
+void lcp_restart(sessionidt s);
 void processlcp(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l);
 void processipcp(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l);
 void processipv6cp(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l);
 void processipin(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l);
+void processmpin(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l);
+void processmpframe(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l, uint8_t extra);
 void processipv6in(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l);
 void processccp(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l);
 void sendchap(sessionidt s, tunnelidt t);
-uint8_t *makeppp(uint8_t *b, int size, uint8_t *p, int l, sessionidt s, tunnelidt t, uint16_t mtype);
+uint8_t *makeppp(uint8_t *b, int size, uint8_t *p, int l, sessionidt s, tunnelidt t, uint16_t mtype, uint8_t prio, bundleidt bid, uint8_t mp_bits);
+uint8_t *opt_makeppp(uint8_t *p, int l, sessionidt s, tunnelidt t, uint16_t mtype, uint8_t prio, bundleidt bid, uint8_t mp_bits);
 void sendlcp(sessionidt s, tunnelidt t);
 void send_ipin(sessionidt s, uint8_t *buf, int len);
 void sendccp(sessionidt s, tunnelidt t);
 void protoreject(sessionidt s, tunnelidt t, uint8_t *p, uint16_t l, uint16_t proto);
+int join_bundle(sessionidt s);
 
 
 // radius.c
@@ -745,7 +947,7 @@ void radiusretry(uint16_t r);
 uint16_t radiusnew(sessionidt s);
 void radiusclear(uint16_t r, sessionidt s);
 void processdae(uint8_t *buf, int len, struct sockaddr_in *addr, int alen, struct in_addr *local);
-
+int rad_tunnel_pwdecode(uint8_t *pl2tpsecret, size_t *pl2tpsecretlen, const char *radiussecret, const uint8_t * auth);
 
 // l2tpns.c
 clockt backoff(uint8_t try);
@@ -753,6 +955,7 @@ void send_ipv6_ra(sessionidt s, tunnelidt t, struct in6_addr *ip);
 void route6set(sessionidt s, struct in6_addr ip, int prefixlen, int add);
 sessionidt sessionbyip(in_addr_t ip);
 sessionidt sessionbyipv6(struct in6_addr ip);
+sessionidt sessionbyipv6new(struct in6_addr ip);
 sessionidt sessionbyuser(char *username);
 void increment_counter(uint32_t *counter, uint32_t *wrap, uint32_t delta);
 void random_data(uint8_t *buf, int len);
@@ -765,13 +968,19 @@ int tun_write(uint8_t *data, int size);
 void adjust_tcp_mss(sessionidt s, tunnelidt t, uint8_t *buf, int len, uint8_t *tcp);
 void sendipcp(sessionidt s, tunnelidt t);
 void sendipv6cp(sessionidt s, tunnelidt t);
-void processudp(uint8_t *buf, int len, struct sockaddr_in *addr);
+void processudp(uint8_t *buf, int len, struct sockaddr_in *addr, uint16_t indexudpfd);
+void processipout(uint8_t *buf, int len);
 void snoop_send_packet(uint8_t *packet, uint16_t size, in_addr_t destination, uint16_t port);
 int find_filter(char const *name, size_t len);
 int ip_filter(uint8_t *buf, int len, uint8_t filter);
-int cmd_show_ipcache(struct cli_def *cli, char *command, char **argv, int argc);
-int cmd_show_hist_idle(struct cli_def *cli, char *command, char **argv, int argc);
-int cmd_show_hist_open(struct cli_def *cli, char *command, char **argv, int argc);
+int cmd_show_ipcache(struct cli_def *cli, const char *command, char **argv, int argc);
+int cmd_show_hist_idle(struct cli_def *cli, const char *command, char **argv, int argc);
+int cmd_show_hist_open(struct cli_def *cli, const char *command, char **argv, int argc);
+tunnelidt lac_new_tunnel();
+void lac_tunnelclear(tunnelidt t);
+void lac_send_SCCRQ(tunnelidt t, uint8_t * auth, unsigned int auth_len);
+void lac_send_ICRQ(tunnelidt t, sessionidt s);
+void lac_tunnelshutdown(tunnelidt t, char *reason, int result, int error, char *msg);
 
 #undef LOG
 #undef LOG_HEX
@@ -791,7 +1000,8 @@ void become_master(void);	// We're the master; kick off any required master init
 
 
 // cli.c
-void init_cli(char *hostname);
+void init_cli();
+void cli_init_complete(char *hostname);
 void cli_do_file(FILE *fh);
 void cli_do(int sockfd);
 int cli_arg_help(struct cli_def *cli, int cr_ok, char *entry, ...);
@@ -802,6 +1012,7 @@ void host_unreachable(in_addr_t destination, uint16_t id, in_addr_t source, uint
 
 
 extern tunnelt *tunnel;
+extern bundlet *bundle;
 extern sessiont *session;
 extern sessionlocalt *sess_local;
 extern ippoolt *ip_address_pool;
@@ -820,16 +1031,19 @@ extern int epollfd;
 
 struct event_data {
 	enum {
-	    	FD_TYPE_CLI,
-	    	FD_TYPE_CLUSTER,
-	    	FD_TYPE_TUN,
-	    	FD_TYPE_UDP,
-	    	FD_TYPE_CONTROL,
-	    	FD_TYPE_DAE,
+		FD_TYPE_CLI,
+		FD_TYPE_CLUSTER,
+		FD_TYPE_TUN,
+		FD_TYPE_UDP,
+		FD_TYPE_CONTROL,
+		FD_TYPE_DAE,
 		FD_TYPE_RADIUS,
 		FD_TYPE_BGP,
+		FD_TYPE_NETLINK,
+		FD_TYPE_PPPOEDISC,
+		FD_TYPE_PPPOESESS
 	} type;
-	int index; // for RADIUS, BGP
+	int index; // for RADIUS, BGP, UDP
 };
 
 #define TIME (config->current_time)
@@ -841,19 +1055,4 @@ extern uint16_t MSS;
 #define CLI_HELP_REQUESTED	(argc > 0 && argv[argc-1][strlen(argv[argc-1])-1] == '?')
 #define CLI_HELP_NO_ARGS	(argc > 1 || argv[0][1]) ? CLI_OK : cli_arg_help(cli, 1, NULL)
 
-// CVS identifiers (for "show version file")
-extern char const *cvs_id_arp;
-extern char const *cvs_id_cli;
-extern char const *cvs_id_cluster;
-extern char const *cvs_id_constants;
-extern char const *cvs_id_control;
-extern char const *cvs_id_icmp;
-extern char const *cvs_id_l2tpns;
-extern char const *cvs_id_ll;
-extern char const *cvs_id_md5;
-extern char const *cvs_id_ppp;
-extern char const *cvs_id_radius;
-extern char const *cvs_id_tbf;
-extern char const *cvs_id_util;
-
 #endif /* __L2TPNS_H__ */
